/* 
 * Copyright (C) 2016-2017 Nexiode
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The complete text of the GNU General Public License
 * can be found in root directory of this program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <stdint.h>

#include <openssl/sha.h>
#include <openssl/aes.h>
#include <openssl/evp.h>
#include <openssl/md5.h>

#include "make-manifest.h"

static struct option long_options[] =
{ { "type",  required_argument, NULL, 't'},
  { "file",  required_argument, NULL, 'f'},
  { NULL,    no_argument,       NULL, 0  },
};

static \
void print_help(char *filename)
{
  fprintf(stderr, "Usage: %s --type <type> --file <filename>\n\n  <type> = CFG, MODBUS or PROV\n", filename);
}

static uint8_t aes_key[AES_KEY_SIZE] = {DEFAULT_AES_KEY};
static uint8_t aes_iv[AES_KEY_SIZE] = {DEFAULT_AES_IV};

/*
 * Function : calc_csum
 * Brief    : Calculate checksum of a given file
 * Param    : Path and filename
 * Return   : Checksum
 */
uint8_t *calc_csum(char *filename)
{
  FILE *fd;
  uint8_t *csum;

  csum = NULL;

  fd = fopen(filename, "rb");

  if (fd == NULL)
  {
    fprintf(stderr, "File '%s' doesn't exist\n", filename);
  }
  else
  {
    MD5_CTX ctx;
    char buf[512];
    int32_t bytes; 

    MD5_Init(&ctx);

    bytes = fread(buf, 1, sizeof(buf), fd);

    while (bytes > 0)
    {
      MD5_Update(&ctx, buf, bytes);

      bytes = fread(buf, 1, sizeof(buf), fd);
    }

    csum = malloc(MD5_DIGEST_LEN);

    if (csum != NULL)
    {
      MD5_Final(csum, &ctx);
    }

    fclose(fd);
  }

  return csum;
}

/*
 * Function : write_manifest
 * Brief    : Write manifest
 * Param    : Filename and manifest
 * Return   : None
 */
static \
void write_manifest(char *filename, char *manifest)
{
  FILE *fd;
  int32_t len;

  fd = fopen(filename, "w");

  if (fd == NULL)
  {
    return;
  }

  len = strlen(manifest);

  if (len > 0)
  {
    fwrite(manifest, 1, len, fd);
  }

  fclose(fd);
}

/*
 * Function : aes_encrypt
 * Brief    : Encrypt data data with AES key
 * Param    : Data to encrypt and length
 * Return   : Ciphered data
 */
static \
uint8_t *aes_encrypt(uint8_t *plaindata, uint32_t *len)
{
  EVP_CIPHER_CTX ctx;
  int32_t c_len = *len + AES_BLOCK_SIZE;
  int32_t f_len = 0;
  uint8_t *cipherdata = malloc(c_len);

  EVP_CIPHER_CTX_init(&ctx);

  EVP_EncryptInit_ex(&ctx, EVP_aes_256_cbc(), NULL, aes_key, aes_iv);

  EVP_EncryptInit_ex(&ctx, NULL, NULL, NULL, NULL);

  EVP_EncryptUpdate(&ctx, cipherdata, &c_len, plaindata, *len);

  EVP_EncryptFinal_ex(&ctx, cipherdata+c_len, &f_len);

  *len = c_len + f_len;

  EVP_CIPHER_CTX_cleanup(&ctx);

  return cipherdata;
}

int main(int argc, char *argv[])
{
  int p;
  char filename[FILENAME_LEN];
  char type_manifest[TYPE_MANIFEST_LEN];
  uint8_t *cs;

  filename[0] = 0;
  type_manifest[0] = 0;

  if (argc == 1)
  {
    print_help(argv[0]);

    return -1;
  }

  while (1)
  {
    p = getopt_long(argc, argv, "f:", long_options, NULL);

    if (p == -1)
    {
      break;
    }

    switch (p)
    {
      case 't':
      {
        strncpy(type_manifest, optarg, TYPE_MANIFEST_LEN);
        type_manifest[TYPE_MANIFEST_LEN-1] = 0;

        break;
      }
      case 'f':
      {
        strncpy(filename, optarg, FILENAME_LEN);
        filename[FILENAME_LEN-1] = 0;

        break;
      }
      default:
      {
        print_help(argv[0]);
        return -1;

        break;
      }
    }
  }

  if ( (strlen(type_manifest) == 0) ||
       ( (strcasecmp(type_manifest, TYPE_MANIFEST_CFG) != 0) &&
         (strcasecmp(type_manifest, TYPE_MANIFEST_MODBUS) != 0) &&
         (strcasecmp(type_manifest, TYPE_MANIFEST_PROV) != 0) ) )
  {
    print_help(argv[0]);

    return -1;
  }

  if (strlen(filename) == 0)
  {
    print_help(argv[0]);

    return -1;
  }

  cs = calc_csum(filename);

  if (cs != NULL)
  {
    uint8_t *encrypted_cs;
    uint32_t len;

    len = MD5_DIGEST_LEN;

    encrypted_cs = aes_encrypt(cs, &len);

    if (encrypted_cs != NULL)
    {
      char filename_manifest[FILENAME_LEN];
      char *manifest;

      manifest = malloc(len*2+1+strlen(filename)+2);

      if (manifest != NULL)
      {
        uint8_t i;
        char *ptr_filename;

        for(i=0; i<len; i++)
        {
          sprintf(&manifest[2*i], "%02x", encrypted_cs[i]);
        }

        manifest[2*len] = 0;

        strcat(manifest, " ");
        strcat(manifest, filename);
        strcat(manifest, "\n");

        if (strcasecmp(type_manifest, TYPE_MANIFEST_CFG) == 0)
        {
          sprintf(filename_manifest, "c_010.manifest");
        }
        if (strcasecmp(type_manifest, TYPE_MANIFEST_MODBUS) == 0)
        {
          sprintf(filename_manifest, "c_modbus_010.manifest");
        }
        if (strcasecmp(type_manifest, TYPE_MANIFEST_PROV) == 0)
        {
          sprintf(filename_manifest, "p_010.manifest");
        }

        write_manifest(filename_manifest, manifest);

        free(manifest);
      }
      free(encrypted_cs);
    }

    free(cs);
  }

  return 0;
}

